-- Plugin Authors: You're welcome to copy, borrow, destroy, improve, set fire to or ignore this code as you see fit.

--get a reference to Hermes
local Hermes = LibStub("AceAddon-3.0"):GetAddon("Hermes")

--setup our addon (AceAddon)
local MODNAME = "LightUI"
local LightUI = Hermes:NewModule(MODNAME, "AceEvent-3.0", "AceTimer-3.0")

--Localization support
local L = LibStub('AceLocale-3.0'):GetLocale("LightUI")

--LSM used for fonts and textures
local LIB_LibSharedMedia = LibStub("LibSharedMedia-3.0") or error("Required library LibSharedMedia-3.0 not found")

--global to local performance improvements
local GetTime = GetTime
local format, floor = format, floor
local tinsert, tremove, pairs, ipairs, wipe, type, tonumber, sort, tostring = tinsert, tremove, pairs, ipairs, wipe, type, tonumber, sort, tostring

------------------------------------------------------------------
--PROPERTIES
------------------------------------------------------------------
local recycler = nil							--contains all the reused bars so that we don't have memory leaks
local anchor = nil								--the drag frame that all bars are parented to
local bars = {}									--the bars currently being drawn

--misc static values used for UI
local ANCHOR_BAR_HEIGHT = 18
local BAR_SPARK_WIDTH = 10
local BAR_SPARK_HEIGHT = 10
local BAR_FONTSPACING = 2
local BAR_GAP = 1
local BAR_BGALPHA = 0.35

local dbp = { }									--the current profile, gets set during OnSetHermesPluginProfile

------------------------------------------------------------------
--UTILITIES
------------------------------------------------------------------
local function _indexOf(tbl, item)				--helper function that locates the index of an item in a table
	for index, i in ipairs(tbl) do
		if(i == item) then
			return index
		end
	end

	return nil
end

local function _removeItem(tbl, item)			--helper function that removes an element in a table
	local index = _indexOf(tbl, item) 
	if not index then error("failed to locate item in table") end
	tremove(tbl, index)
end

local function _secondsToClock(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if nSeconds < 60 then
		--seconds
		return nil, nil, floor(nSeconds)
	else
		if(nSeconds > 3600) then
			--hours
			local nHours = floor(nSeconds / 3600)
			local nMins = floor(nSeconds / 60 - (nHours * 60))
			local nSecs = floor(nSeconds - nHours * 3600 - nMins * 60)
			return nHours, nMins, nSecs
		else
			--minutes
			local nMins = floor(nSeconds / 60)
			local nSecs = floor(nSeconds - nMins * 60)
			return nil, nMins, nSecs
		end
	end
	
	--catch all to avoid any accidental nil errors
	return nil, nil, 0
end

------------------------------------------------------------------
--LIGHTUI MOD
------------------------------------------------------------------
function LightUI:OnEnable()						--called when this addon is enabled (via AceAddon or Hermes callback)
	--provide events for starting and stopping receiving, which is when we'll want to show the UI
	Hermes:RegisterHermesEvent("OnStartReceiving", MODNAME, function() self:OnStartReceiving() end)
	Hermes:RegisterHermesEvent("OnStopReceiving", MODNAME, function() self:OnStopReceiving() end)
end

function LightUI:OnDisable()					--called when the addon is disabled (via AceAddon or Hermes callback)
	Hermes:UnregisterHermesEvent("OnStartReceiving", MODNAME)
	Hermes:UnregisterHermesEvent("OnStopReceiving", MODNAME)
	--kill all the UI stuff
	self:OnStopReceiving()
end

function LightUI:OnInitialize()					--called when this addon is first started (AceAddon)
	LIB_LibSharedMedia.RegisterCallback(self, "LibSharedMedia_Registered", "UpdateUsedMedia") --register for media update callbacks
	
	Hermes:RegisterHermesPlugin(
		MODNAME,
		function () self:Enable() end,
		function () self:Disable() end,
		function (dbplugins) self:OnSetHermesPluginProfile(dbplugins) end,
		function () return self:GetBlizzOptionsTable() end
	)
end

function LightUI:HookEvents(hook)				--sets/unsets hermes events for this plugin
	if hook and hook == true then
		Hermes:RegisterHermesEvent("OnAbilityInstanceStartCooldown", MODNAME, function(instance) self:OnCooldownStarted(instance) end)
		Hermes:RegisterHermesEvent("OnAbilityInstanceUpdateCooldown", MODNAME, function(instance) self:OnCooldownUpdate(instance) end)
		Hermes:RegisterHermesEvent("OnAbilityInstanceStopCooldown", MODNAME, function(instance) self:OnCooldownStopped(instance) end)
		Hermes:RegisterHermesEvent("OnAbilityInstanceRemoved", MODNAME, function(instance) self:OnCooldownRemoved(instance)end)
	else
		Hermes:UnregisterHermesEvent("OnAbilityInstanceStartCooldown", MODNAME)
		Hermes:UnregisterHermesEvent("OnAbilityInstanceUpdateCooldown", MODNAME)
		Hermes:UnregisterHermesEvent("OnAbilityInstanceStopCooldown", MODNAME)
		Hermes:UnregisterHermesEvent("OnAbilityInstanceRemoved", MODNAME)
	end
end

------------------------------------------------------------------
--HERMES PLUGIN CALLBACKS
------------------------------------------------------------------
function LightUI:OnSetHermesPluginProfile(dbplugins) --our handler for when hermes provides us with the profile settings
	if not dbplugins[MODNAME] then
		dbplugins[MODNAME] = self:GetDefaultOptions()
	end
	dbp = dbplugins[MODNAME]
	
	--upgrade options
	if not dbp.hideSelf then
		dbp.hideSelf = false
	end
end

function LightUI:GetBlizzOptionsTable()			--our handler for whne hermes requests our options table
	local options = {
		name = "Light UI",
		type="group",
		childGroups="tree",
		inline = false,
		handler = self,
		args = {
			common = {
				type="group",
				name=" ",
				inline = true,
				order = 5,
				args = {
					general = {
						type="group",
						name="",
						inline = true,
						order = 5,
						args = {
							testbars = {
								type = "execute",
								name = L["Create Test Bars"],
								width = "normal",
								order = 5,
								func = "StartEmulator",
							},
						},
					},
					anchor = {
						type = "group",
						name = L["Window"],
						inline = true,
						order = 10,
						args = {
							locked = {
								type = "toggle",
								name = L["Lock"],
								width = "normal",
								get = function(info) return dbp.locked end,
								order = 3,
								set = function(info, value)
									dbp.locked = value
									self:LockAnchor()
								end,
							},
							growup = {
								type = "toggle",
								name = L["Grow Up"],
								width = "normal",
								get = function(info) return dbp.growup end,
								order = 5,
								set = function(info, value)
									dbp.growup = value
									self:LayoutBars()
								end,
							},
							hideSelf = {
								type = "toggle",
								name = L["Hide Self"],
								width = "normal",
								get = function(info) return dbp.hideSelf end,
								order = 7,
								set = function(info, value)
									dbp.hideSelf = value
									self:LayoutBars()
								end,
							},
							isltor = {
								type = "select",
								name = L["Direction"],
								order = 10,
								style = "dropdown",
								width = "full",
								get = function(info)
									if dbp.isltor == true then
										return "LEFT TO RIGHT"
									else
										return "RIGHT TO LEFT"
									end
								end,
								set = function(info, value)
									if value == "LEFT TO RIGHT" then
										dbp.isltor = true
									else
										dbp.isltor = false
									end
									
									self:ReinitBars()
									self:LayoutBars()
								end,
								values = {
									["LEFT TO RIGHT"] = L["Left to Right"],
									["RIGHT TO LEFT"] = L["Right to Left"],
								},
							},
							scale = {
								type = "range",
								min = 0.1, max = 3, step = 0.01,
								name = L["Scale"],
								order = 15,
								width = "full",
								get = function(info) return dbp.scale end,
								set = function(info, value)
									dbp.scale = value
									anchor:SetScale(dbp.scale)
								end
							},
							alpha = {
								type = "range",
								min = 0, max = 1, step = 0.01,
								name = L["Alpha"],
								order = 20,
								width = "full",
								get = function(info) return dbp.alpha end,
								set = function(info, value)
									dbp.alpha = value
									anchor:SetAlpha(dbp.alpha)
								end
							},
						},
					},
					bars = {
						type = "group",
						name = L["Bars"],
						inline = true,
						order = 15,
						args = {
							height = {
								type = "range",
								min = 1, max = 45, step = 1,
								name = L["Height"],
								order = 5,
								width = "full",
								get = function(info) return dbp.barheight end,
								set = function(info, value)
									dbp.barheight = value
									self:ReinitBars()
									self:LayoutBars()
								end
							},
							width = {
								type = "range",
								min = 50, max = 500, step = 1,
								name = L["Width"],
								order = 10,
								width = "full",
								get = function(info) return dbp.barwidth end,
								set = function(info, value)
									dbp.barwidth = value
									anchor:SetWidth(dbp.barwidth)
									self:ReinitBars()
								end
							},
							bartexture = {
								type = 'select',
								order = 15,
								dialogControl = 'LSM30_Statusbar',
								name = L["Texture"],
								values = AceGUIWidgetLSMlists.statusbar,
								width = "full",
								get = function() return dbp.bartexture end,
								set = function(info, value)
									dbp.bartexture = value
									self:ReinitBars()
								end,
							},
							font = {
								type = "select",
								dialogControl = 'LSM30_Font',
								order = 20,
								name = L["Font"],
								width = "full",
								values = AceGUIWidgetLSMlists.font,
								get = function(info) return dbp.font end,
								set = function(info, value)
									dbp.font = value
									self:ReinitBars()
								end,
							},
							fontsize = {
								type = "range",
								min = 5, max = 30, step = 1,
								name = L["Font Size"],
								width = "full",
								get = function(info) return dbp.fontsize end,
								order = 25,
								set = function(info, value)
									dbp.fontsize = value
									self:ReinitBars()
								end,
							},
							textratio = {
								type = "range",
								min = 0, max = 100, step = 1,
								name = L["Player/Duration Width Ratio (%)"],
								width = "full",
								get = function(info) return dbp.textratio end,
								order = 30,
								set = function(info, value)
									dbp.textratio = value
									self:ReinitBars()
								end,
							},
						},
					},
				},
			},
		},
	}
	
	return options
end

function LightUI:OnStartReceiving()				--our handler for when hermes stops receiving
	self:CreateRecycler()	--won't create it if it's already been created.
	self:CreateAnchor()		--won't create it if it's already been created.
	self:RestoreAnchorPos()
	
	--start capturing events
	self:HookEvents(true)
end

function LightUI:OnStopReceiving()				--our handler for when hermes starts receiving
	--stop capturing instance events
	self:HookEvents(false)
	--shutdown the emulator if it's running
	self:StopEmulator()
	--kill any remaining bars
	while #bars > 0 do
		self:RecycleBar(bars[1])
	end
	--force the anchor to hide itself
	self:LockAnchor(true)
end

------------------------------------------------------------------
--HERMES EVENT CALLBACKS
------------------------------------------------------------------
function LightUI:OnCooldownRemoved(instance)	--our handler for when hermes removes an instance
	local bar = self:GetBarForInstance(instance)
	if bar then
		self:RecycleBar(bar)
		self:LayoutBars()
	end
end

function LightUI:OnCooldownStarted(instance)	--our handler for when hermes starts a cooldown
	local bar = self:GetBarForInstance(instance)
	if not bar then
		--need to create a new bar
		bar = self:FetchBar(instance)
	else
		--bar already exists, do nothing. It will get reinitialized and a redrawn
	end
	
	self:InitBar(bar)
	self:LayoutBars()
end

function LightUI:OnCooldownUpdate(instance)		--our handler for when hermes refreshes a cooldown
	local bar = self:GetBarForInstance(instance)
	if bar then
		self:AnimateBar(bar)
	end
end

function LightUI:OnCooldownStopped(instance)	--our handler for when hermes stops a cooldown
	local bar = self:GetBarForInstance(instance)
	if bar then
		self:RecycleBar(bar)
		self:LayoutBars()
	end
end

------------------------------------------------------------------
--RECYCLER FRAME
------------------------------------------------------------------
function LightUI:CreateRecycler()				--this frame holds resused bars so that we don't cause memory leaks
	if not recycler then
		recycler = CreateFrame("Frame", nil, UIParent)
		recycler:SetPoint("CENTER")
		recycler:SetWidth(50)
		recycler:SetHeight(50)
		recycler:Hide()

		recycler:EnableMouse(false)
		recycler:SetMovable(false)
		recycler:SetToplevel(false)
		recycler.bars = {}
	end
end

------------------------------------------------------------------
--ANCHOR FRAME
------------------------------------------------------------------
local function OnAnchorDragStart(frame, button)	--called when anchor frame starts being dragged
	if not dbp.locked then
		frame:StartMoving()
	end
end

local function OnAnchorDragStop(frame, button)	--called when anchor frame is done being dragged
	if not dbp.locked then
		frame:StopMovingOrSizing()
		LightUI:SaveAnchorPos()
	end
end

function LightUI:SaveAnchorPos()				--updates saved variables for anchor position
	dbp.x = anchor:GetLeft()
	dbp.y = anchor:GetTop()
end

function LightUI:RestoreAnchorPos()				--restores the anchor frame position from saved variables
	if(not dbp.x or not dbp.y) then
		anchor:ClearAllPoints()
		anchor:SetPoint("CENTER", UIParent, "CENTER")
		dbp.x = anchor:GetLeft()
		dbp.y = anchor:GetTop()
	else
		x = dbp.x
		y = dbp.y

		anchor:ClearAllPoints()
		anchor:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
		anchor:SetUserPlaced(nil)
	end
end

function LightUI:CreateAnchor()					--creates the anchor frame
	if not anchor then
		anchor = CreateFrame("Frame", nil, UIParent)
		anchor:SetWidth(dbp.barwidth)
		anchor:SetHeight(ANCHOR_BAR_HEIGHT)
		anchor:EnableMouse(true)
		anchor:SetMovable(true)
		anchor:SetClampedToScreen(true)
		anchor:SetUserPlaced(false)
		anchor:SetScale(dbp.scale)
		anchor:SetAlpha(dbp.alpha)
		anchor:SetPoint("CENTER", UIParent, "CENTER")
		anchor:SetScript("OnDragStart", OnAnchorDragStart)
		anchor:SetScript("OnDragStop", OnAnchorDragStop)
		anchor:RegisterForDrag("LeftButton");
		
		anchor.text = anchor:CreateFontString(nil, "ARTWORK")
		anchor.text:Hide()
		anchor.text:SetFontObject(GameFontNormalSmall);
		anchor.text:SetTextColor(0.9, 0.9, 0.9, 1)
		anchor.text:SetText("Hermes: Light UI")
		anchor.text:SetWordWrap(false)
		anchor.text:SetNonSpaceWrap(false)
		anchor.text:SetJustifyH("CENTER");
		anchor.text:SetJustifyV("CENTER");
		anchor.text:SetAllPoints();
	end

	self:LockAnchor()
end

function LightUI:LockAnchor(force)					--shows or hides the anchor and enabled/disables dragging
	--we don't actually hide the window because everything is parented to it.
	--instead we'll just change the properties to make it appear invisible
	if anchor then
		if dbp.locked == true or force then
			anchor:EnableMouse(false)
			anchor:SetBackdrop(nil)
			anchor:SetBackdropColor(0, 0, 0, 0)
			anchor:SetBackdropBorderColor(0, 0, 0, 0)
			anchor.text:Hide();
		else
			anchor:EnableMouse(true)
			anchor:SetBackdrop({
				bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
				tile = false,
				tileSize = 32,
				edgeSize = 10,
				insets = {
					left = 0,
					right = 0,
					top = 0,
					bottom = 0,
				}
			})
			anchor:SetBackdropColor(0.6, 0.6, 0.6, 0.7)
			anchor:SetBackdropBorderColor(0.6, 0.6, 0.6, 0.7)
			anchor.text:Show();
		end
	end
end

------------------------------------------------------------------
--BARS
------------------------------------------------------------------
function LightUI:GetTextWidths()				--calculates the width for the name and time text
	local maxw = dbp.barwidth - dbp.barheight --height is subtracted due to the icon size
	local w_player, w_duration
	
	--account for 0% player name ratio
	if dbp.textratio <= 0 then
		w_player = 0
		w_duration = maxw
	elseif dbp.textratio >= 100 then
		w_player = maxw
		w_duration = 0
	else
		w_player = maxw * (dbp.textratio / 100)
		w_duration = maxw - w_player
	end
	
	return w_player, w_duration
end

function LightUI:GetBarForInstance(instance)	--searches for the bar matching a given hermes instance
	for _, bar in ipairs(bars) do
		if bar.instance == instance then
			return bar
		end
	end
end

function LightUI:LayoutBars()					--sorts and repositions the bars
	local height = dbp.barheight
	local width = dbp.barwidth
	
	--go ahead and resort the bars based on duration
	self:BarSort()
	
	--now reposition all the bars
	local count = 1
	for _, bar in ipairs(bars) do
		if dbp.hideSelf and dbp.hideSelf == true and bar.instance.sender.name == UnitName("player") then
			--hide the bar
			bar:Hide()
			bar:ClearAllPoints()
		else
			local offset, apBar, apAnchor
			offset = ( count - 1 ) * (height + BAR_GAP)
			if dbp.growup == false then
				--down
				offset = offset * -1
				apBar = "TOP"
				apAnchor = "BOTTOM"
			else
				--up
				apBar = "BOTTOM"
				apAnchor = "TOP"
			end

			bar:ClearAllPoints()
			bar:SetPoint(apBar, anchor, apAnchor, 0, offset) -- TOP/BOTTOM assignment
			bar:SetPoint("LEFT", anchor, "LEFT", 0, 0)
			bar:SetPoint("RIGHT", anchor, "RIGHT", 0, 0)
			bar:SetHeight(height)
			
			count = count + 1
			
			--show the bar
			bar:Show()
		end
	end
end

function LightUI:AnimateBar(bar)				--updates the duration text and bar animations
	local maxw = dbp.barwidth - dbp.barheight --height is subtracted due to the icon size
	local width, sparkpos
	
	--it might seem dumb to not just format the string inside of _secondsToClock
	--but the reason I do it this way is because no memory is allocated for strings
	--using this technique. Also, the CPU required is reduced by roughly 1/3rd
			
	local h, m, s = _secondsToClock(bar.instance.remaining)
	if h then
		bar.time:SetFormattedText("%d:%02.f:%02.f", h, m, s)
	elseif m then
		bar.time:SetFormattedText("%d:%02.f", m, s)
	else
		bar.time:SetFormattedText("%d", s)
	end
	
	if dbp.isltor == true then
		width = (bar.instance.remaining / bar.instance.initialDuration) * maxw
		--offset by the icon width too
		sparkpos = dbp.barheight + width
	else
		width = (bar.instance.remaining / bar.instance.initialDuration) * maxw
		sparkpos = (maxw - width)
	end
	
	bar.fg:SetWidth(width)
	
	bar.spark:SetPoint("CENTER", bar, "LEFT", sparkpos, 0)
	bar.spark:SetHeight(dbp.barheight + BAR_SPARK_HEIGHT)
	bar.spark:SetWidth(BAR_SPARK_WIDTH)
	bar.spark:Show()
end

function LightUI:ReinitBars()					--resets values based on config changes
	for _, bar in ipairs(bars) do
		local bartexture = LIB_LibSharedMedia:Fetch("statusbar", dbp.bartexture)
		local font = LIB_LibSharedMedia:Fetch("font", dbp.font)
		
		bar.bg:SetTexture(bartexture)
		bar.fg:SetTexture(bartexture)
		
		local wp, wd = LightUI:GetTextWidths()
		
		--lt
		bar.name:SetFont(font, dbp.fontsize, nil)
		
		--rt
		bar.time:SetFont(font, dbp.fontsize, nil)
		
		bar.bg:ClearAllPoints()
		bar.fg:ClearAllPoints()
		bar.icon:ClearAllPoints()
		bar.name:ClearAllPoints()
		bar.time:ClearAllPoints()
		bar.spark:ClearAllPoints()
		
		--set points for icon, text, and textures
		if dbp.isltor == true then 
			--left to right (icon on the left)
			bar.bg:SetPoint("TOPLEFT", bar, "TOPLEFT", dbp.barheight, 0)
			bar.bg:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", dbp.barheight, 0)
			bar.bg:SetWidth(dbp.barwidth - dbp.barheight)
			
			bar.fg:SetPoint("TOPLEFT", bar, "TOPLEFT", dbp.barheight, 0)
			bar.fg:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", dbp.barheight, 0)
			
			bar.icon:SetPoint("TOPLEFT", bar, "TOPLEFT", 0, 0)
			bar.icon:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", 0, 0)
			bar.icon:SetWidth(dbp.barheight)
			
			bar.name:SetPoint("TOPLEFT", bar.icon, "TOPRIGHT", BAR_FONTSPACING, 0)
			bar.name:SetPoint("BOTTOMLEFT", bar.icon, "BOTTOMRIGHT", BAR_FONTSPACING, 0)
			bar.name:SetJustifyH("LEFT");
			
			if wp > 0 then
				bar.name:SetWidth(wp)
			end
			
			bar.time:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -BAR_FONTSPACING, 0)
			bar.time:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -BAR_FONTSPACING, 0)
			bar.time:SetJustifyH("RIGHT");
			
			if wd > 0 then
				bar.time:SetWidth(wd)
			end
		else
			--right to left (icon on the right)
			--bg
			bar.bg:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -dbp.barheight, 0)
			bar.bg:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -dbp.barheight, 0)
			bar.bg:SetWidth(dbp.barwidth - dbp.barheight)
			
			--fg
			bar.fg:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -dbp.barheight, 0)
			bar.fg:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -dbp.barheight, 0)
			
			--icon
			bar.icon:SetPoint("TOPRIGHT", bar, "TOPRIGHT", 0, 0)
			bar.icon:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", 0, 0)
			bar.icon:SetWidth(dbp.barheight)
			
			--lt
			bar.name:SetPoint("TOPRIGHT", bar.icon, "TOPLEFT", -BAR_FONTSPACING, 0)
			bar.name:SetPoint("BOTTOMRIGHT", bar.icon, "BOTTOMLEFT", -BAR_FONTSPACING, 0)
			bar.name:SetJustifyH("RIGHT");
			
			if wp > 0 then
				bar.name:SetWidth(wp)
			end
			
			--rt
			bar.time:SetPoint("TOPLEFT", bar, "TOPLEFT", BAR_FONTSPACING, 0)
			bar.time:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", BAR_FONTSPACING, 0)
			bar.time:SetJustifyH("LEFT");
			
			if wd > 0 then
				bar.time:SetWidth(wd)
			end
		end
	end
end

function LightUI:InitBar(bar)					--sets all necessary first time values
	local playerclass = bar.instance.sender.class
	local playername = bar.instance.sender.name
	local spellname = bar.instance.ability.name
	local spellicon = bar.instance.ability.icon
	local barcolor = Hermes:GetClassColorRGB(playerclass)
	local bartexture = LIB_LibSharedMedia:Fetch("statusbar", dbp.bartexture)
	local font = LIB_LibSharedMedia:Fetch("font", dbp.font)
	
	--bg
	bar.bg:SetTexture(bartexture)
	bar.bg:SetVertexColor(0.3, 0.3, 0.3, BAR_BGALPHA)
	bar.bg:ClearAllPoints()
	
	--fg
	bar.fg:SetTexture(bartexture)
	bar.fg:SetVertexColor(barcolor.r, barcolor.g, barcolor.b, 1)
	bar.fg:ClearAllPoints()
	
	--icon
	bar.icon:SetTexture(spellicon)
	bar.icon:ClearAllPoints()
	
	local wp, wd = LightUI:GetTextWidths()
	
	--lt
	bar.name:SetFont(font, dbp.fontsize, nil)
	bar.name:SetTextColor(1, 1, 1, 1)
	bar.name:SetText(playername)
	bar.name:SetShadowColor(0, 0, 0, 0.5)
	bar.name:SetShadowOffset(1, -1)
	bar.name:ClearAllPoints()
	
	--rt
	bar.time:SetFont(font, dbp.fontsize, nil)
	bar.time:SetTextColor(1, 1, 1, 1)
	bar.time:SetShadowColor(0, 0, 0, 0.5)
	bar.time:SetShadowOffset(1, -1)
	bar.time:SetText(nil)
	bar.time:ClearAllPoints()
	
	--spark
	bar.spark:ClearAllPoints()
	bar.spark:Hide()
	
	--set points for icon, text, and textures
	if dbp.isltor == true then 
		--left to right (icon on the left)
		bar.bg:SetPoint("TOPLEFT", bar, "TOPLEFT", dbp.barheight, 0)
		bar.bg:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", dbp.barheight, 0)
		bar.bg:SetWidth(dbp.barwidth - dbp.barheight)
		
		bar.fg:SetPoint("TOPLEFT", bar, "TOPLEFT", dbp.barheight, 0)
		bar.fg:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", dbp.barheight, 0)
		bar.fg:SetWidth(dbp.barwidth - dbp.barheight)
		
		bar.icon:SetPoint("TOPLEFT", bar, "TOPLEFT", 0, 0)
		bar.icon:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", 0, 0)
		bar.icon:SetWidth(dbp.barheight)
		
		bar.name:SetPoint("TOPLEFT", bar.icon, "TOPRIGHT", BAR_FONTSPACING, 0)
		bar.name:SetPoint("BOTTOMLEFT", bar.icon, "BOTTOMRIGHT", BAR_FONTSPACING, 0)
		bar.name:SetJustifyH("LEFT");
		
		if wp > 0 then
			bar.name:SetWidth(wp)
		end
		
		bar.time:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -BAR_FONTSPACING, 0)
		bar.time:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -BAR_FONTSPACING, 0)
		bar.time:SetJustifyH("RIGHT");
		
		if wd > 0 then
			bar.time:SetWidth(wd)
		end
	else
		--right to left (icon on the right)
		bar.bg:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -dbp.barheight, 0)
		bar.bg:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -dbp.barheight, 0)
		bar.bg:SetWidth(dbp.barwidth - dbp.barheight)
		
		bar.fg:SetPoint("TOPRIGHT", bar, "TOPRIGHT", -dbp.barheight, 0)
		bar.fg:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", -dbp.barheight, 0)
		bar.fg:SetWidth(dbp.barwidth - dbp.barheight)
		
		bar.icon:SetPoint("TOPRIGHT", bar, "TOPRIGHT", 0, 0)
		bar.icon:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", 0, 0)
		bar.icon:SetWidth(dbp.barheight)
		
		bar.name:SetPoint("TOPRIGHT", bar.icon, "TOPLEFT", -BAR_FONTSPACING, 0)
		bar.name:SetPoint("BOTTOMRIGHT", bar.icon, "BOTTOMLEFT", -BAR_FONTSPACING, 0)
		bar.name:SetJustifyH("RIGHT");
		
		if wp > 0 then
			bar.name:SetWidth(wp)
		end
		
		bar.time:SetPoint("TOPLEFT", bar, "TOPLEFT", BAR_FONTSPACING, 0)
		bar.time:SetPoint("BOTTOMLEFT", bar, "BOTTOMLEFT", BAR_FONTSPACING, 0)
		bar.time:SetJustifyH("LEFT");
		
		if wd > 0 then
			bar.time:SetWidth(wd)
		end
	end
end

function LightUI:CreateBar()					--bars are created only once
	--bar frame
	local bar = CreateFrame("Frame", nil, recycler)
	bar:Hide() -- hide by default
	bar:ClearAllPoints()
	bar.instance = nil
	bar:EnableMouse(false) --allow clickthrough
	bar:SetMovable(false)
	bar:SetResizable(false)
	bar:SetToplevel(true)
	
	--bar bg
	bar.bg = bar:CreateTexture(nil, "BACKGROUND")
	bar.bg:ClearAllPoints()
	
	--bar fg
	bar.fg = bar:CreateTexture(nil, "BORDER")
	bar.fg:ClearAllPoints()

	--bar icon
	bar.icon = bar:CreateTexture(nil, "OVERLAY")
	bar.icon:ClearAllPoints()
	bar.icon:SetTexture("")
	
	--bar lt
	bar.name = bar:CreateFontString(nil, "ARTWORK", "GameFontWhite")
	bar.name:ClearAllPoints()
	bar.name:SetWordWrap(false)
	bar.name:SetJustifyV("CENTER");
	bar.name:SetJustifyH("LEFT");
	bar.name:SetText("")
	bar.name:SetTextColor(1, 1, 1, 1)
	
	--bar rt
	bar.time = bar:CreateFontString(nil, "ARTWORK", "GameFontWhite")
	bar.time:ClearAllPoints()
	bar.time:SetWordWrap(false)
	bar.time:SetJustifyV("CENTER");
	bar.time:SetJustifyH("RIGHT");
	bar.time:SetText("")
	bar.time:SetTextColor(1, 1, 1, 1)

	--spark for internal cooldowns
	bar.spark = bar:CreateTexture(nil, "OVERLAY")
	bar.spark:ClearAllPoints()
	bar.spark:SetTexture("Interface\\CastingBar\\UI-CastingBar-Spark")
	bar.spark:SetBlendMode("ADD")
	
	tinsert(recycler.bars, bar)
	
	return bar
end

function LightUI:FetchBar(instance)				--gets a new bar or pulls one from the recycler
	local bar
	if(#recycler.bars > 0) then
		bar = recycler.bars[1]
	else
		bar = self:CreateBar()
	end

	_removeItem(recycler.bars, bar)
	tinsert(bars, bar)
	
	bar.instance = instance
	bar:SetParent(anchor)
	return bar
end

function LightUI:RecycleBar(bar)				--recycles an existing bar
	_removeItem(bars, bar)
	tinsert(recycler.bars, bar)
	
	---------------------------------------------------------------
	-- BAR
	---------------------------------------------------------------
	bar:Hide()
	bar.instance = nil
	bar.bg:SetTexture("")
	bar.bg:ClearAllPoints()
	bar.fg:SetTexture("")
	bar.fg:ClearAllPoints()
	bar.icon:SetTexture("")
	bar.icon:ClearAllPoints()
	bar.name:SetText("")
	bar.name:ClearAllPoints()
	bar.time:SetText("")
	bar.time:ClearAllPoints()
	bar:SetParent(recycler)
	bar:ClearAllPoints()
end

------------------------------------------------------------------
--SORTING
------------------------------------------------------------------
function LightUI:BarSort()						--very simple sort function that sorts by duration
	table.sort(bars,
		function(a, b)
			return a.instance.remaining > b.instance.remaining
		end
	)
end

------------------------------------------------------------------
--OPTIONS
------------------------------------------------------------------
function LightUI:GetDefaultOptions()			--default options for new profiles
	local defaults = {
		locked = false,
		scale = 1,
		alpha = 1,
		barwidth = 180,
		barheight = 18,
		bartexture = "Blizzard",
		textratio = 60,
		growup = false, --true = down, false = up
		isltor = true, --left to right = true, right to left = false
		font = "Friz Quadrata TT",
		fontsize = 12,
		hideSelf = false,
	}
	
	return defaults
end

------------------------------------------------------------------
-- HERMES EMULATOR

-- This shows you a simple way to emulate Hermes. In this case,
-- we are emulating Hermes in order to show some test bars to
-- help the user setup their UI
------------------------------------------------------------------
local emulator = nil							--if not nil, then the emulator is running
local TEST_INSTANCES = {						--these model the table structure that Hermes uses for callbacks containing "instance"
	{
		testduration = 60,
		ability = {
			["name"] = "Aura Mastery",
			["icon"] = "Interface\\Icons\\Spell_Holy_AuraMastery",
		},
		sender = {
			["name"] = "Shrimp",
			["class"] = "PALADIN",
		},
	},
	{
		testduration = 120,
		ability = {
			["name"] = "Guardian Spirit",
			["icon"] = "Interface\\Icons\\Spell_Holy_GuardianSpirit",
		},
		sender = {
			["name"] = "Gump",
			["class"] = "PRIEST",
		},
	},
	{
		testduration = 180,
		ability = {
			["name"] = "Heroism",
			["icon"] = "Interface\\Icons\\Ability_Shaman_Heroism",
		},
		sender = {
			["name"] = "Bubba",
			["class"] = "SHAMAN",
		},
	},
}

local function OnEmulatorUpdate(frame, elapsed)	--update all the bars being emulated
	emulator.lastscan = emulator.lastscan + elapsed;
	while (emulator.lastscan > emulator.freq) do
		emulator.lastscan = emulator.lastscan - emulator.freq
		for _, instance in ipairs(emulator.instances) do
			--update remaining time
			instance.remaining = instance.initialDuration - (GetTime()- instance.initialTimeStamp)
			
			if instance.remaining <= 0 then
				--item just went off cooldown
				instance.remaining = nil
				instance.initialDuration = nil
				instance.initialTimeStamp = nil

				--fire off that it's done
				LightUI:OnCooldownStopped(instance) --simulate Hermes callback
				
				--remove instance from emulator
				_removeItem(emulator.instances, instance)
				
				--stop the emulator if there are no more test instances
				if #emulator.instances == 0 then
					LightUI:StopEmulator()
					return
				end
			else
				--item still on cooldown
				LightUI:OnCooldownUpdate(instance) --simulate Hermes callback
			end
		end
	end
end

function LightUI:StartEmulator()				--start the emulator, fires off a bar for each TEST_INSTANCE
	--stop the emulator if it's running
	self:StopEmulator()

	--setup the emulator
	emulator = { lastscan = 0, freq = 0.05, instances = {} }
	
	--add spells to the emulator
	for _, instance in ipairs(TEST_INSTANCES) do
		--reset the instance properties
		instance.initialDuration = instance.testduration
		instance.remaining = instance.testduration
		instance.initialTimeStamp = GetTime()
		
		--add instance to test instances
		tinsert(emulator.instances, instance)
		
		--fire off a bar
		self:OnCooldownStarted(instance) --simulate Hermes callback
	end
	
	--start the emulator
	anchor:SetScript("OnUpdate", OnEmulatorUpdate)
end

function LightUI:StopEmulator()					--stop the emulator, clears all bars and resets emulator
	if emulator then
		--stop events
		anchor:SetScript("OnUpdate", nil)
		
		--stop remaining bars, if any
		for _, instance in ipairs(emulator.instances) do
			self:OnCooldownStopped(instance)
		end
		
		wipe(emulator)
		emulator = nil
	end
end

------------------------------------------------------------------
--LSM Support
------------------------------------------------------------------
function LightUI:UpdateUsedMedia(event, mediatype, key)
	--update all the cells in the container
	if mediatype == "statusbar" or mediatype == "font" then
		LightUI:ReinitBars()
	end
end
